#ifndef __assessTime__
#define __assessTime__
#include "config.h"
#include "system.h"
#include "rtl.h"
#include "basic-block.h"
#include "flags.h"
#include "tree.h"
#include <stdlib.h>
#include <math.h>
#undef false
#undef true


#define my_n_bb (n_basic_blocks * 5)
/* basic blocks can be deleted but the bb no's are not changed. So we
 * cannot just assume that each bb no is < n_basic_blocks.
 */

#define add_global_timeout {            \
        cond[0].kind = "guard";         \
        cond[0].pred = "T_s0 <= T_o";   \
}

#define myassert(cond, error_msg){             \
        if (!(cond)){                                         \
                fprintf(stderr, "\nError at line no %d: %s\n", \
                                __LINE__, error_msg);    \
                abort();                                   \
        }                                                  \
}

#define BBID(bb) ((bb)->index >= 0 ? (bb)->index:my_n_bb+DELTA + (bb)->index)
#define BBINDX(idx) ((idx) >= 0 ? (idx):(my_n_bb+DELTA + (idx)))
#define T_1 2929857
#define T_U 20
#define TIME_OUT 6816965


#define maxTime "max_time"
#define MATCH_FOUND 1
#define FOUND 1
#define NOT_FOUND 0
#define AUTODIFF 500
#define MAX_FUNC_LEN 50
#define MAX_FUNC_CNT 1000
#define NC (automata*AUTODIFF)
#define NCI(a) ((a)*AUTODIFF)
#define FOR_EACH_AUTOMATA for( automata=0;automata < BUF_SIZE; ++automata)
#define FOR_ALL_BB_IN_WL(wl,b) if ((wl)!= NULL) \
                for(b=(wl)->bb;wl!=NULL;wl=wl->next,b=(wl==NULL?NULL:wl->bb))
#define LEAF_NEXT(bbid) (AUTODIFF/2+(bbid))
#define LEAF_NEXT2(bbid) (3*AUTODIFF/4+(bbid))

#define print_RS(fout) fprintf(fout, DELIM"\n");

#define CALL_COST 20
/* May need to flush the inst cache. */

#define TTL (999999999/BASE)
/* Will not die for most purposes. */
/* Need to do more -- add more constraints, if we want to use TTL as a
 * parameter. TTL = time to livej
 */

#define ZERO (AUTODIFF-11)
#define FIRST (AUTODIFF-12)
#define ERR1 (AUTODIFF-14)
#define ERR2 (AUTODIFF-15)
#define TEMP1 (AUTODIFF-16)
/* Make sure the relative values of TOSTATE and TOSTATE2 does not change.
 * We use TOSTATE and TOSTATE2 to emualte the timeout wait.
 */
#define TOSTATE (AUTODIFF-17)
#define TOSTATE2 (AUTODIFF-18)
#define XN1 1
#define XN2 2
#define XN3 3
#define A0 1
#define A2 2
#define A0A1 4
#define A1A2 5
#define A2A3 6
#define A1 7
#define A3 8

#define DELTA 10
/* DELTA gives the possible number of basic blocks that have negative
 * index! I do not know how they are negative in the first place. Anyways.
 * In this new frame of model, there would not be any need for this. So
 * setting it to 0. Can be removed.
 */
#define BAD_NODE -1
#define WAIT_INCR 10
#define MAX_INTR_CNT 1000
#define CUR_FNAME        IDENTIFIER_POINTER (DECL_NAME (current_function_decl))
#define NEEDS_CARE 0
#define CARE_TAKEN 1
#define SUCCESS 0
#define FAILURE 1
#define BASE 1
#define FACTOR 1

#define CYCLE_FREE 1
#define CYCLE_PRESENT 2

#define PROCESS_BEGIN "my_process_start"
#define PROCESS_END "my_process_end"
#define SYN_RECV_START "my_syn_receive_start"
#define SYN_RECV_END "my_syn_receive_end"
#define FIRST_AUTOMATA 0

typedef enum Bool {
        false=0, 
        true=1
} boolean;
typedef struct FList  {
        const char *fname;
/*        // Function name.*/
        int count;
/*        // Number of times the function fname is called.*/
        struct FList *next;
/*        // Pointer to the rest of the list.*/

} FuncList;
typedef struct TVal {
        long long int int_cost;
/*        // int_cost represents a valid value only if it is nonzero. */
        FuncList *func_cost;
/*        // func_cost represents a valid value only if it is non NULL.*/
} timeVal;

typedef struct condition {
        const char *kind;
        const char *pred; /* predicate */
}Cond ;

typedef struct Trans{
        int src;
        int dest;
        Cond *cond;
        int ncond;
} Transition;

typedef struct node_condition{
        Cond *cond;
        int ncond;
} NodeCond;

struct BB_Cost{
        long long int cost;
        long long int mcost; /* max cost */
        boolean back_edge_head;
}*bb_cost;

typedef struct Worklist {
        basic_block bb;
        struct Worklist *next;
} worklist;

typedef struct Call_Graph_And_Result {
       long long int cost;
       boolean marker_present;
       struct Call_Graph_And_Result *marker_cgr;
       /* If marker_cgr is null => no marker found. Else it contains the
        * call graph and result for the path START -> MARKER
        */
       /* We are interested in the following costs with respect to
        * markers. 
        * START -> MARKER1
        * Currerntly, I am taking care of one marker only. There is only
        * one marker procedure.
        */

} CGR;


void do_init(void);
void add_to_timeVal(timeVal *dest, timeVal *new_val);
timeVal * make_timeVal_from_int(long long int);
timeVal * make_timeVal_from_func(FuncList *);
FuncList* create_flist(const char *,int, FuncList *);
FuncList* add_to_flist(FuncList *, char *fname);
FuncList* add_Flists(FuncList *, FuncList *);
timeVal * create_timeVal(void);
/*// Ensures that no duplicates are added.*/
#define delete_from_flist myassert(false, "Delete operation not allowed")
void deallocate_list(FuncList *);
void deallocate_timeVal(timeVal *tv);
void *mycalloc(int ,int);
/*// Called only after the analysis is over or for debugging purposes.*/
void print_cost(timeVal *);

boolean useless_insn(rtx);
boolean call_insn(rtx);
const char *Fname(rtx insn);
long long int Farg(rtx insn);
int get_frequency(basic_block bb);
void update_with_frequency(timeVal *tv, int freq);
long long int already_known(const char *fname);

int add_transitions(int bb1, int bb2, int choice);
void set_node_condition(int bb_no, Cond *cond,int ncond);
Cond *ND_cond(int bb_no);
int ND_ncond(int bb_no);
void do_nothing(long long int t_o,long long int timer_period, int
                m, long long int t_1, long long int t_u);
void create_edge(int src, int dest, Cond* cond, int ncond);
void create_and_print_node(int bb_no, const char *node_name, Cond *cond, int ncond);
void set_init(int nd_no);
void end_template(void);
int read_known_costs(void);
void prologue(void);
void attacker(void);
void epilogue(long long int t_o,long long int  timer_period, int
                m, long long int t_1, long long int t_u);
void begin_template(void);
void do_time_assessment(void);
void print_edge(int bb_s, int bb_d, Cond *cond,int ncond);
void print_conditions(Cond *cond,int);
boolean bb_has_useful_inst(basic_block bb);

int assess_time(void);
worklist *make_wl(basic_block);
long long do_DFT(basic_block , char *);
CGR* do_BFT(worklist *);
CGR* BFT_for_marker(worklist *,int);
extern int global_freq_max;
void set_cost(int, long long int);
long long int compute_cost(basic_block);
void set_max_cost(int, long long int);
void add_bb_to_wl(worklist **, basic_block);

boolean visited(int);
void reset_visited(int);
void set_visited(int);
void deallocate_wl(worklist*);
void deallocate_visited_flag(void);
long long int return_bb_max_cost(int bb);
long long int return_bb_cost(int bb);
void add_ncond_to_edge(int bb1,int bb2, Cond* cond, int ncond);
void add_cond_to_edge(int bb1,int bb2, Cond* cond);
void add_cond_to_node(int bb, int automata, Cond *cond);
void copy_cond (Cond *dest, Cond *src);
worklist * push_conditions_on_path(basic_block bb1, basic_block bb2, int automata, Cond *cond,int ncond, worklist *);
void add_natural_condition(int bb, int automata);
boolean neighbors(basic_block bb1, basic_block bb2);
int check_presence(worklist *wl, basic_block bb);
boolean is_annotation(const char *fname);
boolean is_loop_head(int bb);
void set_loop(int bb);
void build_atc(int bb);
void add_timeout_edges(void);
void timer_intr_handler(void);
int can_loose_intr(int bb);
boolean invalid_bb(basic_block bb);
boolean is_leaf(basic_block bb);
int make_timer_interaction_cond(int noc, Cond **argCnd);
int make_loop_cond(Cond **);
void make_edges_for_end_of_processing_to_loop_back(
                int src1, 
                int src2,
                int dest1,
                int dest2,
                int dest3,
                boolean last_automata);
void timer_default_increment(int bb,int automata, Cond *cond);
void print_CGR(FILE *fout, CGR *cgr);
void create_guard(Cond *cond, const char * const s);
void create_assignment(Cond *cond, const char * const s);
void create_synchronization(Cond *cond, const char * const s);
void print_wl(worklist *wl);
extern int tdiCount;


long long int get_func_cost(const char * const fname);

extern int __LINENO__;
#endif
